// ********************************************************************
// EQ Prozessor (mit RC5 Infrarotdecoder)
// Stand: 16.10.2008
// SDCC-Version: M-IDE-51/SDCC 2.8.0
// Mikrocontroller: P89LPC922 (8051-Derivat mit 2 Clock Core)
// SDCC Aufruf-Parameter: SDCC eq_rc5.c --code-loc 0x0000 --iram-size 256
// M-IDE-51 Compiler-Einstellung: --code-loc 0x0000 --iram-size 256
// ********************************************************************
// Fr NXP-8051-Microcontroller, P89LPC922
// Frequenz: Intern 7.3728 MHz
// Reset-Pin: Resetfunktion deaktiviert (FlashMagic)
// Code-Speicherbedarf: ca. 7kByte
// (Hinweis: Im Bereich 0x1E00 - 0x1FFF liegt der intere Factory-Bootloader)
// Das Programm wurde mit dem P89LPC922 getestet.
// ********************************************************************
// Die Programmfunktionen drfen fr private Nutzung verwendet und
// verndert werden. Fr gewerbliche Nutzung ist die Zustimmung
// des Authors erforderlich. Die Nutzung erfolgt auf eigene Gefahr.
// Fr eventuell entstehende Schden wird keine Haftung bernommen.
// Author: Matthias Kahnt - mx2003@gmx.net
// ********************************************************************

#include "P89LPC922.h"   // Register des P89LPC922
#include <string.h>      // Stingfunktionen

// Typendefinition
typedef unsigned char byte;
typedef unsigned int  word;
typedef unsigned long dword;

// Definebereich
#define LED2        P1_7
#define LED1        P1_6
#define TASTE_ESC   P1_5
#define TASTE_SETM  P0_1
#define TASTE_SETP  P0_3
#define TASTE_ET    P0_6

// I2C-Bus
#define SDA         P1_3
#define SCL         P1_2
#define	ACK         0
#define	NACK        1
#define ADD_7317    0x84

// TDA 7317 Grundinitialisierung
#define INIT_PEGEL 0x2F
#define INIT_BAND1 0x80
#define INIT_BAND2 0x90
#define INIT_BAND3 0xA0
#define INIT_BAND4 0xB0
#define INIT_BAND5 0xC0

// Infrarotempfang RC5
#define INIT_RC5_ADDR   5
#define INIT_RC5_ESC    1
#define INIT_RC5_SETM   2
#define INIT_RC5_SETP   3
#define INIT_RC5_ET     6
#define RC5_DATA     P1_4
#define RC5_L           6

// LC-Display
#define SPALTEN        16
#define ZEILEN          3
#define HGL_ZEIT       60
#define LED2_ZEIT       2
#define MENUE_TIMEOUT 120
#define INIT_LEVEL      2
#define INIT_KONTRAST  16
#define LCD_RS       P3_0
#define LCD_SCL      P1_0
#define LCD_DATA     P1_1
#define LCD_CSB      P3_1
#define LCD_HGL      P0_0

// LPC Datenflash
#define LOAD         0x00
#define ERASEPROGRAM 0x68
#define FLASHADDR    0x0080
#define FLASHLEN     15

// ab Adresse FLASHADDR im Code-Bereich FLASHLEN Byte reservieren
// Hinweis: SDCC legt Programm-Code erst dahinter an(?)
code at FLASHADDR byte dataflash[FLASHLEN] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

// Timer 0 - Zeitbasis
// Timer0: Reload Wert fr T0 T=1/(fosz/2)*(65536-(TH0*256+TL0))
// RELOAD (16bit) = -(0.01 * FREQ_OSC / 2 - 65536);
// 200us = 0xFD1E bei FOSZ 7.373000 MHz
#define TH0Reload 0xFD
#define TL0Reload 0x1E

// Globale Variablen
static byte str0[20];
static byte idata lcd_kontrast1;
static byte idata lcd_kontrast2;
static bit i2c_error          = 0;
volatile byte ad_pol          = 0;
volatile byte ad_por          = 0;
volatile bit mess_pol         = 0;
volatile bit mess_por         = 0;
volatile bit mess_zyklus      = 0;
volatile word timer_tick      = 5000;
volatile byte hgl_counter     = 0;
volatile byte led2_counter    = 0;
volatile byte timeout_counter = 0;
static byte rc5_level_tick    = 0;
static dword rc5_temp         = 0;
static dword rc5_bit_pos      = 0;
static bit rc5_level          = 0;
static bit rc5_zuvor          = 0;
static bit rc5_pol            = 0;
static bit rc5_error          = 0;
volatile bit rc5_ready        = 0;
volatile bit rc5_zyklus       = 0;

// Struktur der Anwender- und Systemeinstellungen, Speicherung im Datenflash
struct awd_struktur {  
  byte kennb;    // Kennbyte fr Flashdaten gltig
  byte pegel;    // Grundverstrkung des TDA 7317
  byte band1;    // Band 1    60Hz
  byte band2;    // Band 2   260Hz
  byte band3;    // Band 3  1030Hz
  byte band4;    // Band 4  3150Hz
  byte band5;    // Band 5 10300Hz
  byte level;    // Aussteuerungsanzeige (Teiler)
  byte kontrast; // LCD Kontrast
  byte rc5addr;  // RC5 Adresse
  byte rc5esc;   // RC5 Code fr ESC Taste
  byte rc5setm;  // RC5 Code fr SET- Taste
  byte rc5setp;  // RC5 Code fr SET+ Taste
  byte rc5et;    // RC5 Code fr ET Taste
};
struct awd_struktur idata awd; 

// Struktur der per Infrarot-FB empfangenen Daten
struct rc5_struc {
  byte adresse;  // RC5 Adresse
  byte daten;    // RC5 Daten
  byte toggle;   // RC5 Toggle-Bit
  byte toggle1;  // RC5 Toggle-Bit der zuvor empfangenen Daten
};
struct rc5_struc idata rc5;   

// Tabelle 1 fr 8 selbst definierte Sonderzeichen (Querstriche)
byte code sonderzeichen1[] = {
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x1F, // 0 (Strich ganz unten)
  0x00,0x00,0x00,0x00,0x00,0x00,0x1F,0x00, // 1
  0x00,0x00,0x00,0x00,0x00,0x1F,0x00,0x00, // 2
  0x00,0x00,0x00,0x00,0x1F,0x00,0x00,0x00, // 3
  0x00,0x00,0x00,0x1F,0x00,0x00,0x00,0x00, // 4
  0x00,0x00,0x1F,0x00,0x00,0x00,0x00,0x00, // 5
  0x00,0x1F,0x00,0x00,0x00,0x00,0x00,0x00, // 6
  0x1F,0x00,0x00,0x00,0x00,0x00,0x00,0x00  // 7 (Strich ganz oben)
};

// Tabelle 2 fr 8 selbst definierte Sonderzeichen (Lngsstriche)
byte code sonderzeichen2[] = {
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 0 Leer
  0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10, // 1 (Strich ganz links)
  0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18, // 2 (2 Striche links)
  0x1C,0x1C,0x1C,0x1C,0x1C,0x1C,0x1C,0x1C, // 3 (3 Striche links)
  0x1E,0x1E,0x1E,0x1E,0x1E,0x1E,0x1E,0x1E, // 4 (4 Striche links)
  0x1F,0x1F,0x1F,0x1F,0x1F,0x1F,0x1F,0x1F, // 5 (5 Striche links)
  0x00,0x1F,0x10,0x17,0x14,0x17,0x14,0x17, // 6 E
  0x00,0x1F,0x01,0x1D,0x09,0x09,0x09,0x09  // 7 T
};

// ********************************************************************
// Timer 0 Interrupt
// Zeit: 200us Intervall
// ********************************************************************
void TIMER0ISR (void) interrupt 1 {
  TH0 = TH0Reload; // Update MSB
  TL0 = TL0Reload; // Update LSB
  // Analogwert-Erfassung fr die Level-Anzeige
  if (mess_zyklus) {
    TH1--;   // PWM Funktion zur Erzeugung der Rampenfunktion
    // Abfrage, ob POR bzw. POL die Rampe geschnitten hat
    if ((mess_por) && (!(CMP1 & 0x02))) { ad_por = 0xFF - TH1; mess_por = 0; }
    if ((mess_pol) && (!(CMP2 & 0x02))) { ad_pol = 0xFF - TH1; mess_pol = 0; }
    if (!TH1) mess_zyklus = 0;
  }
  // Zeitsteuerung fr automatische Funktionen
  timer_tick--; // Zhlt pro Sekunde von 5000 rckwrts bis auf 0
  if (!timer_tick) {
    // Funktion die jede Sekunde 1x ausgefhrt wird
    if (hgl_counter)  hgl_counter--;
      else LCD_HGL = 0; // LCD-Hintergrundbeleuchtung ausschalten
    if (led2_counter) led2_counter--;
      else LED2 = 1;    // LED2 (bersteuerung) ausschalten
    if (timeout_counter) timeout_counter--;
    timer_tick = 5000;
  }
  // RC5 IR-Decodierung
  if (rc5_zyklus) {
    rc5_level_tick++; // Anzahl der Timerdurchlufe fr jeweiligen Pin-Zustand
    if (rc5_level_tick > (RC5_L * 2)) {
      // Fehlimpulse, zu lang, Zyklus beenden
      rc5_zyklus = 0;
      rc5_error  = 1;
    } else {
      // Impulslnge noch nicht zu lang
      rc5_level = RC5_DATA;
      if (rc5_zuvor != rc5_level) {
        // Flankenwechsel, 1. Wechsel ist L/H Flanke (Ende 1. Startbit)
        if (rc5_level_tick < (RC5_L / 2)) {
          // Fehlimpulse, zu kurz, Zyklus beenden
          rc5_zyklus = 0;
          rc5_error  = 1;
        } else {
          // Impulslnge stimmt
          if (!rc5_pol) { // 0=L/H 1=H/L Flanke
            // L/H Flanke
            if (rc5_level_tick <= RC5_L) {
              // letzte Zeit war kurz L
              rc5_bit_pos >>= 1; //0 bernehmen
            } else {
              // letzte Zeit war lang LL
              rc5_bit_pos >>= 2; //00 bernehmen
            }
          } else {
            // H/L Flanke
            if (rc5_level_tick <= RC5_L) {
                // letzte Zeit war kurz H
                rc5_temp |= rc5_bit_pos; // 1 bernehmen
                rc5_bit_pos >>= 1;
              } else {
                // letzte Zeit war lang HH
                rc5_temp |= rc5_bit_pos; // 11 bernehmen
                rc5_bit_pos >>= 1;
                rc5_temp |= rc5_bit_pos;
                rc5_bit_pos >>= 1;
            }
          }
          rc5_pol = !rc5_pol; // nchste Flanke mu andere Polaritt haben
          rc5_level_tick = 0;
        }
        rc5_zuvor = rc5_level;
        if (rc5_bit_pos & 0x00000010) {
          // Alle 28 (Halb)Bits liegen vor
          // Empfangene Bits nun den Datenbytes zuordnen
          if (rc5_temp & 0x08000000) rc5.toggle = 1; else rc5.toggle = 0;
          if (rc5_temp & 0x02000000) rc5.adresse += 16;
          if (rc5_temp & 0x00800000) rc5.adresse +=  8;
          if (rc5_temp & 0x00200000) rc5.adresse +=  4;
          if (rc5_temp & 0x00080000) rc5.adresse +=  2;
          if (rc5_temp & 0x00020000) rc5.adresse +=  1;
          if (rc5_temp & 0x00008000) rc5.daten   += 32;
          if (rc5_temp & 0x00002000) rc5.daten   += 16;
          if (rc5_temp & 0x00000800) rc5.daten   +=  8;
          if (rc5_temp & 0x00000200) rc5.daten   +=  4;
          if (rc5_temp & 0x00000080) rc5.daten   +=  2;
          if (rc5_temp & 0x00000020) rc5.daten   +=  1;
          LED2       = 0; // LED 2 einschalten
          rc5_zyklus = 0; // keine weiteren Daten erfassen, Zyklus beenden
          rc5_ready  = 1; // Signalisieren, dass ein RC5 Datensatz vorliegt
        }
      }
    }
    if (rc5_error) EX1 = 1; // Fehlerfall, INTR1 wieder freigeben und auf neuen IR-Impulse warten
  }
}

// ***********************************************************************
// INTR1 (P1.4 RC5) Interruptroutine
// H/L-Flanke lst diesen Interrupt aus
// (Der TSOP1736 liefert die RC5-Signale invertiert!)
// Impulsfolge am P1.4 Pin:
//
//  ' Bit 1 ' Bit 2 ' Bit 3 ' Bit 4 ' Bit 5 ' usw.
// -----     ---     ---         -------      ---
//      |   |   |   |   |       |       |    |
//       ---     ---     -------         ---
//      '
//      Hier INTR1-Auslsung
//
//      Aufbau eines RC5-Datenblocks:
//      Bit  1: Startbit Nr. 1 (immer 1)
//      Bit  2: Startbit Nr. 2 (immer 1)
//      Bit  3: Togglebit (1 oder 0, im Bsp. oben 1)
//      Bit  4: Adressbit A4 (im Bsp. oben 0)
//      Bit  5: Adressbit A3 (im Bsp. oben 1)
//      Bit  6: Adressbit A2
//      Bit  7: Adressbit A1
//      Bit  8: Adressbit A0
//      Bit  9: Datenbit D5
//      Bit 10: Datenbit D4
//      Bit 11: Datenbit D3
//      Bit 12: Datenbit D2
//      Bit 13: Datenbit D1
//      Bit 14: Datenbit D0
//
// Die Erfassung des RC5-Datenblocks erfolgt im TIMER0-Interrupt!
// ***********************************************************************
void INTR1ISR() interrupt 2 {
  EX1 = 0;         // INTR1 sperren
  rc5.adresse = 0; // RC5 Datensatz lschen
  rc5.daten   = 0;
  rc5.toggle  = 0;
  rc5_temp    = 0x80000000; // Zwischenspeicher der Bitfolge, 28 (Halb)Bits
  rc5_bit_pos = 0x40000000; // Zeiger der empfangenen Bits
  rc5_level_tick = 0; // Anzahl der Timerdurchlufe fr konstanten Pegel am Pin
  rc5_pol     = 0; // Flankenpolaritt 0=L/H 1=H/L Flanke
  rc5_zuvor   = 0; // stellt Pegel der ersten L-Periode dar, die INTR1 ausgelst hat
  rc5_ready   = 0; // Flag fr kompletten rc5 Datensatz empfangen
  rc5_error   = 0; // Flag fr Fehlimpuls
  rc5_zyklus  = 1; // Erfassungszyklus im Timer0 freigeben
}

// ********************************************************************
// Verzgerungsschleife ohne Timer fr kurze Zeit
// Zeit: Wartezeit in [(wert - 1) * 2 + 5]
//       1 = 5us
//       2 = 7us
//       3 = 9us u.s.w. (gilt fr 7.373MHz und LPC9xx MC)
//     255 = 0.5ms
// ********************************************************************
void delay_us(byte zeit) {
byte zaehler;
  for ( zaehler = zeit; zaehler; zaehler-- ) {
   _asm nop; _endasm; // Keil: _nop_();
  }
}

// ********************************************************************
// Verzgerungsschleife ohne Timer fr lngere Zeit
// Zeit: Wartezeit in [ms]
//       1 = ca. 1ms
//       2 = ca. 2ms
//       3 = ca. 3ms u.s.w.
//     100 = ca. 100ms
//    1000 = ca. 1s
// ********************************************************************
void delay(word zeit) {
word zaehler;
  for (zaehler = zeit; zaehler; zaehler--) {
    delay_us(255);  // dauert ca. 0.5ms
    delay_us(255);  // dauert ca. 0.5ms
  }
}

// ********************************************************************
// I2C-Bus
// Startsequenz
// ********************************************************************
void I2C_start(void) {
  SDA = 1; SDA = 1; SDA = 1; SDA = 1; // Mehrfachanweisungen nur aus Timinggrnden
  SCL = 1; SCL = 1; SCL = 1; SCL = 1; // dto.
  SDA = 0; SDA = 0; SDA = 0; SDA = 0; // ...
}

// ********************************************************************
// I2C-Bus
// Stoppsequenz
// ********************************************************************
void I2C_stop(void) {
  SDA = 0; SDA = 0; SDA = 0; SDA = 0;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  SDA = 1; SDA = 1; SDA = 1; SDA = 1;
}

// ********************************************************************
// I2C-Bus
// Write: 1 Byte schreiben
// bergabe: senddat - zu bertragendes Zeichen
// ********************************************************************
void I2C_write(byte senddat) {
byte i;
  SCL = 0; SCL = 0; SCL = 0; SCL = 0;
  for (i = 1; i <= 8; i++) { // 8 Bit seriell senden
   SDA = senddat >> 7;
   SCL = 1; SCL = 1; SCL = 1; SCL = 1;
   senddat <<= 1;
   SCL = 0; SCL = 0; SCL = 0; SCL = 0;
  }
  SDA = 1; // Portpin SDA fr Empfang das ACK-Bits vorbereiten
  SCL = 0; SCL = 0; SCL = 0; SCL = 0;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  if (SDA) i2c_error = 1; // ACK erwartet, aber nicht erhalten
  SCL = 0; SCL = 0; SCL = 0; SCL = 0;
}

// ***********************************************************************
// Schreiben eines Zeichens an das LCD-Modul (seriell Soft-SPI Mode)
// bergabe: lcd_byte : Auszugebendes Zeichen/Steuerzeichen
//           lcd_mode : 0 - Daten
//                      1 - Steuerzeichen
// ***********************************************************************
void write_lcd(byte lcd_byte, bit lcd_mode) {
byte stelle;
  // Pause bevor nchstes Zeichen gesendet werden darf
  delay_us(20); // Pause mind. 26,3us laut DOG-M Spezifikation
  LCD_CSB = 0;  // LCD-Modul freigeben
  if (lcd_mode) LCD_RS = 0; // Steuerzeichen
    else LCD_RS = 1;        // Daten
  // Byte senden seriell senden, MSB zuerst (D7 D6 D5 D4 D3 D2 D1 D0)
  for ( stelle = 0x80; stelle; stelle >>= 1 ) {
    LCD_DATA = lcd_byte & stelle;
    LCD_SCL = 0; LCD_SCL = 0;
    LCD_SCL = 1; LCD_SCL = 1;
  }
  LCD_CSB = 1; // LCD-Modul sperren
}

// ***********************************************************************
// Lschen des LC-Display
// und Kursor auf Position 0,0 setzen
// ***********************************************************************
void clear_lcd(void) {
  write_lcd(0x01,1);
  delay(30); // Wartezeit, lschen dauert einige ms
} 

// ***********************************************************************
// Ausgabe eines ASCII-Zeichen positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN-1)
//           lcd_y : Zeile  (0...ZEILEN-1)
//           lcd_ascii : ASCII-Zeichen
// *********************************************************************** 
void write_byte_lcd(byte lcd_x, byte lcd_y, byte lcd_ascii) {
byte lcd_offset = 0;
  if (lcd_x > (SPALTEN - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break; // Zeile 1
    case 1:  lcd_offset = 0x10; break; // Zeile 2 
    case 2:  lcd_offset = 0x20; break; // Zeile 3
  };
  write_lcd(0x80 + lcd_x + lcd_offset,1); // Kursorposition setzen
  write_lcd(lcd_ascii,0);                 // Ausgabe des ASCII-Zeichen an der Kursorposition
}

// ***********************************************************************
// Ausgabe einer Zeichenkette positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN-1)
//           lcd_y : Zeile  (0...ZEILEN-1)
//           lcd_zeichen : Adresse der auszugebenden format.  Zeichenkette
//           clr_line    : Lschen bis Zeilenende
//                         1 - Lschen
//                         0 - kein Lschen
// ***********************************************************************
void printf_lcd(byte lcd_x, byte lcd_y, byte *lcd_zeichen, bit clr_line) {
byte lcd_i;
byte lcd_offset = 0;
  if (lcd_x > (SPALTEN - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break; // Zeile 1
    case 1:  lcd_offset = 0x10; break; // Zeile 2 
    case 2:  lcd_offset = 0x20; break; // Zeile 3
  }
  write_lcd(0x80 + lcd_x + lcd_offset,1); // Kursorposition setzen
  // Ausgabe der Zeichenkette ab der Kursorposition
  lcd_offset = strlen(lcd_zeichen); // Lnge der Zeichenkette
  if (lcd_offset > SPALTEN) lcd_offset = SPALTEN;
  for(lcd_i = lcd_offset; lcd_i; lcd_i--) {
    write_lcd(*lcd_zeichen,0); lcd_zeichen++;
  }
  if (clr_line) {
    // Lschen bis Zeilenende
    for(lcd_i = SPALTEN - lcd_offset - lcd_x; lcd_i; lcd_i--)
      write_lcd(' ',0);
  }
}

// ***********************************************************************
// Blinkposition/Cursor auf dem LC-Display setzen
// bergabe: lcd_x : Spalte (0...SPALTEN-1)
//           lcd_y : Zeile  (0...ZEILEN-1)
//           lcd_blink :   0 - Blinken/Cursor aus
//                         1 - Blinken an
//                         2 - Cursor  an
// ***********************************************************************
void blink_lcd(byte lcd_x, byte lcd_y, byte lcd_blink) {
byte lcd_offset = 0;
  write_lcd(0x0C,1); // KURSOR ausschalten
  if (lcd_x > (SPALTEN - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break; // Zeile 1
    case 1:  lcd_offset = 0x10; break; // Zeile 2 
    case 2:  lcd_offset = 0x20; break; // Zeile 3
  };
  write_lcd(0x80 + lcd_x + lcd_offset,1); // Blinkposition setzen
  if (lcd_blink == 1) write_lcd(0x0D,1);  // Blinken ein
  if (lcd_blink == 2) write_lcd(0x0E,1);  // Cursor ein
}

// ***********************************************************************
// Definieren eines Sonderzeichen auf dem LC-Display
// bergabe: lcd_addr : Adresse
//           lcd_zeichen: Zeiger auf das 1. Byte der Zeichendefinition
// ***********************************************************************
void write_lcd_cg(byte lcd_addr, byte *lcd_zeichen) {
byte lcd_i;
  write_lcd(0x38,1); // Function Set DL=1 N=1 DH=0 IS2=0 IS1=0 IS Table 0
  for(lcd_i = 0; lcd_i < 8; lcd_i++) {
    write_lcd(0x40 + lcd_addr * 8 + lcd_i,1); // CG RAM Adresse Set (01aa azzz)
    write_lcd(*lcd_zeichen,0);  // Data Write 8x Pixelzeile
    lcd_zeichen++;
  }
  write_lcd(0x39,1); // Function Set DL=1 N=1 DH=0 IS2=0 IS1=1 IS Table 1
} 

// ***********************************************************************
// Definieren von 8 Sonderzeichen
// bergabe:  lcd_zeichen: Zeiger auf das Zeichenfeld (8x8 Byte)
// ***********************************************************************
void write_lcd_cg_block(byte *sonderzeichenblock) {
byte lcd_i;
  for(lcd_i = 0; lcd_i < 8; lcd_i++) {
    write_lcd_cg(lcd_i, &sonderzeichenblock[lcd_i * 8]);
  }
} 

// ***********************************************************************
// Formatierung einer Zeichenkette
// ersetzt die System-sprintf-Funktion (verbraucht zu viel Speicher!)
// bergabe: ze_buffer : Adresse des Buffers
//           f_wert    : zu formatierendes Byte
//           f_fkt     : Formatierungsfunktion
// ***********************************************************************
void sprintf_lcd(byte *ze_buffer, byte f_wert, byte f_fkt) {
  byte code HEX_TAB[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
  byte z_i;
  switch (f_fkt) {
    case 1: // Byte in DEZ 2-stellig mit fhrender 0 (00 ... 99)
            *ze_buffer++ = (f_wert / 10) + 0x30;          
            *ze_buffer++ = (f_wert % 10) + 0x30;          
            *ze_buffer   = 0;
            break;
    case 2: // Byte in DEZ von 0 ... 14 dB(2er Schritte, 2-stellig mit Vorzeichen in Bit 3)
            if (f_wert & 0x08) {
              // negativ
              *ze_buffer++ = '-';          
              *ze_buffer++ = (~(f_wert - 1) & 0x07) * 2 / 10 + 0x30;          
              *ze_buffer++ = (~(f_wert - 1) & 0x07) * 2 % 10 + 0x30;          
            } else {
              // positiv oder Null
              if (!f_wert) *ze_buffer++ = ' '; else *ze_buffer++ = '+';          
              *ze_buffer++ = (f_wert & 0x07) * 2 / 10 + 0x30;          
              *ze_buffer++ = (f_wert & 0x07) * 2 % 10 + 0x30;          
            }
            *ze_buffer++ = 'd';          
            *ze_buffer++ = 'B';
            *ze_buffer   = 0;
            break;          
    case 3: // Byte in HEX 2-stellig mit fhrender 0
            *ze_buffer++ = HEX_TAB[f_wert >> 4];          
            *ze_buffer++ = HEX_TAB[f_wert & 0x0F];          
            *ze_buffer   = 0;
            break;
    case 4: // Pegel als Bandanzeige
            // Linker Kanal  Buffer-Adresse 0 ... 7
            // rechter Kanal Buffer-Adresse 8 ... 15
            f_wert = f_wert / awd.level; // Levelanzeige skalieren
            if (f_wert > 1) {
              hgl_counter = HGL_ZEIT; // LCD-Hintergrundbeleuchtung Automatik Ein
              LCD_HGL = 1; // LCD-Hintergrundbeleuchtung Ein
            }
            if (f_wert > 35) {
              // bersteuerung
              f_wert = 35; // Bandanzeige begrenzen
              led2_counter = LED2_ZEIT; // Led2 Automatik ein (bersteuerung)
              LED2 = 0; // Led2 ein
            }
            // Bufferbereich lschen
            for(z_i = 0; z_i < 7; z_i++) *ze_buffer++ = ' ';
            *ze_buffer   = 0; // Endekennung
            ze_buffer   -= 7; // Bufferzeiger wieder auf Anfang setzen
            // Bandanzeige formatieren
            if (!f_wert) break;
            if (f_wert < 6) {
              *ze_buffer++ = f_wert;   // Level 1 ... 5
              break;
            } else *ze_buffer++ = 5;   // Level > 5
            if ((f_wert > 5) && (f_wert < 11)) {
              *ze_buffer++ = f_wert - 5; // Level 6 ... 10
              break;
            } else *ze_buffer++ = 5;   // Level > 10
            if ((f_wert > 10) && (f_wert < 16)) {
              *ze_buffer++ = f_wert - 10; // Level 11 ... 15
              break;
            } else *ze_buffer++ = 5;   // Level > 15
            if ((f_wert > 15) && (f_wert < 21)) {
              *ze_buffer++ = f_wert - 15; // Level 16 ... 20
              break;
            } else *ze_buffer++ = 5;   // Level > 20
            if ((f_wert > 20) && (f_wert < 26)) {
              *ze_buffer++ = f_wert - 20; // Level 21 ... 25
              break;
            } else *ze_buffer++ = 5;   // Level > 25
            if ((f_wert > 25) && (f_wert < 31)) {
              *ze_buffer++ = f_wert - 25; // Level 26 ... 30
              break;
            } else *ze_buffer++ = 5;   // Level > 30
            if ((f_wert > 30) && (f_wert < 36)) {
              *ze_buffer++ = f_wert - 30; // Level 31 ... 35
              break;
            } else *ze_buffer++ = 5;   // Level > 35
  }
}

// ***********************************************************************
// Prfroutine, Taste gedrckt bzw. IR-Tastencode empfangen
// Rckgabe:	1,2,3,4 = Taste gedrckt
//              10 + RC5 Daten mit gltiger Adresse
//              0 = keine Taste gedrckt
// keine Auswertung, wenn mehrere Tasten gleichzeitig gedrckt sind
// ***********************************************************************
byte keypressed(void) {
byte key_code = 0;
  if (!TASTE_ESC)  key_code = 1; // Taste  ESC gedrckt
  if (!TASTE_SETM) key_code = 2; // Taste SET- gedrckt
  if (!TASTE_SETP) key_code = 3; // Taste SET+ gedrckt
  if (!TASTE_ET)   key_code = 4; // Taste   ET gedrckt
  if (rc5_ready) {
    // Infrarot RC5 Daten liegen vor
    if (awd.rc5addr == rc5.adresse) {
      // Adresse stimmt
      if (rc5.toggle1 != rc5.toggle) {
        // Tastendruck nur auswerten, wenn Taste nur eimal gedrckt
        key_code = 10 + rc5.daten; // RC5 Daten zuweisen
        rc5.toggle1 = rc5.toggle;  // Merker fr gleiche Taste dauergedrckt
      }
    }
    rc5_ready = 0;
    EX1 = 1; // RC5 IR Empfang wieder freigeben
  }
  return (key_code);
}

// ***********************************************************************
// Tastatur-Abfrage
// Warten auf Tastendruck und Rckgabe des Kodes der gedrckten Taste
// (IR-Tasten werden auf die Kodes der normalen Tasten umgerechnet)
// Abbruch: System-Timeout
// Rckgabe:	0x01 = Taste ESC
//              0x02 = Taste SET-
//              0x03 = Taste SET+
//              0x04 = Taste ET
//              0x05 = Timeout
// keine Auswertung, wenn mehrere Tasten gleichzeitig gedrckt sind
// ***********************************************************************
byte get_taste(void) {
byte zw_code;
byte key_code;
  zw_code = 0;
  hgl_counter = HGL_ZEIT; // LCD-Hintergrundbeleuchtung Automatik Ein
  LCD_HGL     = 1;        // LCD-Hintergrundbeleuchtung Ein
  timeout_counter = MENUE_TIMEOUT;
  while (1) {
    while(keypressed());  // Warten falls Taste noch nicht losgelassen
    delay(50);
    key_code = 0;
    // Warten auf Tastendruck oder IR-Empfang
    while(!key_code && timeout_counter) {
      key_code = keypressed();
    }
    // Eine Taste wurde gedrckt, oder IR-Daten empfangen, oder Timeout eingetreten
    delay(50); // kurz warten, falls es nur prellen war, dann Taste ermitteln
    if (!TASTE_ESC)       zw_code = 1; // Taste  ESC  gedrckt
    if (!TASTE_SETM)      zw_code = 2; // Taste  SET- gedrckt
    if (!TASTE_SETP)      zw_code = 3; // Taste  SET+ gedrckt
    if (!TASTE_ET)        zw_code = 4; // Taste    ET gedrckt
    if (key_code > 9) {
      // RC5 Daten per IR-Fernbedienung
      key_code -= 10;
      if (key_code == awd.rc5esc)  zw_code = 1; // ESC
      if (key_code == awd.rc5setm) zw_code = 2; // SET-
      if (key_code == awd.rc5setp) zw_code = 3; // SET+
      if (key_code == awd.rc5et)   zw_code = 4; // ET
      //EX1 = 1; // RC5 IR Empfang wieder freigeben
    }
    if (!timeout_counter) zw_code = 5; // Timeout eingetreten
    if (zw_code) break;
  }
  hgl_counter = HGL_ZEIT; // LCD-Hintergrundbeleuchtung Automatik Ein
  LCD_HGL     = 1;        // LCD-Hintergrundbeleuchtung Ein
  return (zw_code);
}

// ********************************************************************
// In Application Programming (IAP-Funktion)
// Byte in Datenflash-Bereich schreiben
// Die Routine nutzt die IAP-FLASH Funktion der LPC900 Familie,
// im Page-Mode (Pageregister-Gre 64-Byte).
// Whrend des Programmierzustandes ist das System automatisch im IDLE-Mode.
// Es drfen keine Interrupts auftreten. Der Vorgang dauert ca. 4ms.
// 100.000 Schreibzyklen sind max. spezifiziert.
// bergabe: dataaddr   - Flashadresse
//           databyte   - Datenbyte
// Rckgabe: 0 - Fehler
//           1 - Ok
// ********************************************************************
bit write_byte_flash(word dataaddr, byte databyte) { 
  FMCON  = LOAD;                             // Load-Kommando
  FMADRL = (byte) (dataaddr & 0x00FF);       // Flash-Adresse Low Byte
  FMADRH = (byte) (dataaddr >> 8);           // Flash-Adresse High Byte
  FMDATA = databyte;                         // Datenbyte
  FMCON  = ERASEPROGRAM;                     // Lsch-/Programmier-Kommando
  if (FMCON & 0x8F) return 0; else return 1; // Fehlerprfung Fehler=0 ok=1
}

// ********************************************************************
// TDA 7317 und System-Einstellungen in Datenflash speichern
// ab Adresse FLASHADDR
// ********************************************************************
void write_data_flash(void) {
  EA = 0; // Interrupts sperren
  awd.kennb = 0x19; // kennzeichnet spter, dass die FLASH-Werte gltig sind
  write_byte_flash((word) &dataflash[0],awd.kennb);
  write_byte_flash((word) &dataflash[1],awd.pegel);
  write_byte_flash((word) &dataflash[2],awd.band1);
  write_byte_flash((word) &dataflash[3],awd.band2);
  write_byte_flash((word) &dataflash[4],awd.band3);
  write_byte_flash((word) &dataflash[5],awd.band4);
  write_byte_flash((word) &dataflash[6],awd.band5);
  write_byte_flash((word) &dataflash[7],awd.level);
  write_byte_flash((word) &dataflash[8],awd.kontrast);
  write_byte_flash((word) &dataflash[9],awd.rc5addr);
  write_byte_flash((word) &dataflash[10],awd.rc5esc);
  write_byte_flash((word) &dataflash[11],awd.rc5setm);
  write_byte_flash((word) &dataflash[12],awd.rc5setp);
  write_byte_flash((word) &dataflash[13],awd.rc5et);
  EA = 1; // Interrupts freigeben
}

// ********************************************************************
// TDA 7317 und System-Einstellungen aus Datenflash lesen
// ab Adresse FLASHADDR
// ********************************************************************
void read_data_flash(void) {
  awd.kennb    = dataflash[0];
  awd.pegel    = dataflash[1];
  awd.band1    = dataflash[2];
  awd.band2    = dataflash[3];
  awd.band3    = dataflash[4];
  awd.band4    = dataflash[5];
  awd.band5    = dataflash[6];
  awd.level    = dataflash[7];
  awd.kontrast = dataflash[8];
  awd.rc5addr  = dataflash[9];
  awd.rc5esc   = dataflash[10];
  awd.rc5setm  = dataflash[11];
  awd.rc5setp  = dataflash[12];
  awd.rc5et    = dataflash[13];
}

// ***********************************************************************
// Equalizer Einstellungen
// ***********************************************************************
void equalizer(void) {
byte band = 0; // 0...4 entspricht Band 1...5 (Index fr aktuelles EQ-Band)
char temp[5];  // 0...4 entspricht Band 1...5 (temporre EQ-Werte)
byte taste;
byte z_i;
  // Equalizer-Werte in vorzeichenbehaftete Zahl berneheme (B7 = Vorzeichen 1=negativ)
  temp[0] = awd.band1 & 0x08 ? (~awd.band1 << 4) + 0x80 : awd.band1 << 4; // Band 1
  temp[1] = awd.band2 & 0x08 ? (~awd.band2 << 4) + 0x80 : awd.band2 << 4; // Band 2
  temp[2] = awd.band3 & 0x08 ? (~awd.band3 << 4) + 0x80 : awd.band3 << 4; // Band 3
  temp[3] = awd.band4 & 0x08 ? (~awd.band4 << 4) + 0x80 : awd.band4 << 4; // Band 4
  temp[4] = awd.band5 & 0x08 ? (~awd.band5 << 4) + 0x80 : awd.band5 << 4; // Band 5
  write_lcd_cg_block(sonderzeichen1); // Sonderzeichen definieren
  clear_lcd();
  printf_lcd(0,0,"Band:  1 2 3 4 5",1);
  printf_lcd(0,1," SET+",1);
  printf_lcd(0,2," SET-",1);
  while(1) {
    // Anzeige der 5 Equalizer-Werte
    for(z_i = 0; z_i < 5; z_i++) {
      if (temp[z_i] < 0) {
        // Abschwchung (Cut)
        write_byte_lcd(7 + z_i * 2,2,(temp[z_i] >> 4) & 0x07);
        write_byte_lcd(7 + z_i * 2,1,' ');
      } else {
        // Anhebung (Boost)
        write_byte_lcd(7 + z_i * 2,1,temp[z_i] >> 4);
        write_byte_lcd(7 + z_i * 2,2,' ');
      }
    }
    blink_lcd(band * 2 + 7,0,1); // aktuelle Einstellposition blinken
    taste = get_taste(); // Tastendruck abwarten
    delay(50);
    if (taste == 2) if (temp[band] > -112) temp[band] -= 16; // Equalizer-Wert verkleinern
    if (taste == 3) if (temp[band] <  112) temp[band] += 16; // Equalizer-Wert vergrern
    if ((taste == 1) || (taste == 5)) break; // ESC oder Timeout
    // neuen Equalizer-Wert zum TDA 7317 bertragen
    I2C_start();            // Startsequenz
    I2C_write(ADD_7317);    // Baustein fr schreiben adressieren
    if (temp[band] >= 0) I2C_write(0x80 + band * 16 + (temp[band] >> 4));  // Band x Boost
      else I2C_write(0x80 + band * 16 + ((~temp[band] >> 4) & 0x0F) + 9);  // Band x Cut
    I2C_stop();             // Stoppsequenz
    if (taste == 4) { // Taste ET (Bandwechsel)
      if (band < 4) band++; else band = 0;
    }
    // dB Anzeige aktualisieren
    printf_lcd(0,1,"     ",0);
    printf_lcd(0,2,"     ",0);
    sprintf_lcd(&str0[0],temp[band] >> 4,2);
    printf_lcd(0,1,str0,0);
  }
  // neue Equalizer-Werte speichern
  awd.band1 = temp[0] >= 0 ? 0x80 + (temp[0] >> 4) : 0x80 + ((~temp[0] >> 4) & 0x0F) + 8;
  awd.band2 = temp[1] >= 0 ? 0x90 + (temp[1] >> 4) : 0x90 + ((~temp[1] >> 4) & 0x0F) + 8;
  awd.band3 = temp[2] >= 0 ? 0xA0 + (temp[2] >> 4) : 0xA0 + ((~temp[2] >> 4) & 0x0F) + 8;
  awd.band4 = temp[3] >= 0 ? 0xB0 + (temp[3] >> 4) : 0xB0 + ((~temp[3] >> 4) & 0x0F) + 8;
  awd.band5 = temp[4] >= 0 ? 0xC0 + (temp[4] >> 4) : 0xC0 + ((~temp[4] >> 4) & 0x0F) + 8;
  write_data_flash(); // Equalizer-Werte auch im Daten-FLASH speichern
  write_lcd_cg_block(sonderzeichen2); // ursprngliche Sonderzeichen wieder definieren
}

// ***********************************************************************
// Pegel Einstellungen
// ***********************************************************************
void pegel(void) {
byte temp;
byte taste;
  temp = (awd.pegel >> 3) * 3; 
  clear_lcd();
  printf_lcd(0,0,"Verst\x84rkung:",1);
  printf_lcd(0,1," SET+",1);
  printf_lcd(0,2," SET-",1);
  while(1) {
    // Anzeige der Pegel
    printf_lcd(7,1,"-   dB",1);
    sprintf_lcd(&str0[0],temp,1);
    printf_lcd(8,1,str0,0);
    blink_lcd(8,1,1);    // aktuelle Einstellposition blinken
    taste = get_taste(); // Tastendruck abwarten
    delay(50);
    if (taste == 2) if (temp > 2)  temp -= 3; // Pegel-Wert verkleinern
    if (taste == 3) if (temp < 15) temp += 3; // Pegel-Wert vergrern
    if ((taste == 1) || (taste == 4) || (taste == 5)) break; // ESC oder ET oder Timeout
    awd.pegel = (temp / 3) << 3;
    // neuen Pegel-Wert zum TDA 7317 bertragen
    I2C_start();            // Startsequenz
    I2C_write(ADD_7317);    // Baustein fr schreiben adressieren
    I2C_write(awd.pegel);   // Pegelwert bertragen
    I2C_stop();             // Stoppsequenz
  }
  write_data_flash(); // Werte auch im Daten-FLASH speichern
}

// ***********************************************************************
// Levelanzeige Einstellungen
// ***********************************************************************
void level(void) {
byte taste;
  clear_lcd();
  printf_lcd(0,0,"Levelanzeige:",1);
  printf_lcd(0,1," SET+",1);
  printf_lcd(0,2," SET-",1);
  while(1) {
    // Anzeige der Einstellung
    printf_lcd(7,1,"\xF8",1);
    sprintf_lcd(&str0[0],awd.level,1);
    printf_lcd(8,1,str0,0);
    blink_lcd(8,1,1); // aktuelle Einstellposition blinken
    taste = get_taste(); // Tastendruck abwarten
    delay(50);
    if (taste == 2) if (awd.level > 1)  awd.level--; // Level-Wert verkleinern
    if (taste == 3) if (awd.level < 15) awd.level++; // Level-Wert vergrern
    if ((taste == 1) || (taste == 4) || (taste == 5)) break; // ESC oder ET oder Timeout
  }
  write_data_flash(); // Werte auch im Daten-FLASH speichern
}

// ********************************************************************
// LCD Kontrast Set
// ********************************************************************
void lcd_kontrast_set (void) {
  // LCD Kontrast setzen
  lcd_kontrast1 &= 0xFC; // 1111 1100   alten Kontrast C5 C4 lschen
  lcd_kontrast1 |= (awd.kontrast >> 4) & 0x03; // neuen Kontrast C5 C4 setzen
  lcd_kontrast2 &= 0xF0; // 1111 0000   alten Kontrast C3 C2 C1 C0 lschen
  lcd_kontrast2 |= awd.kontrast & 0x0F; // neuen Kontrast C3 C2 C1 C0 setzen
  write_lcd(lcd_kontrast1,1); // Power Control + Kontrast Set C5,C4 zur LCD bertragen
  write_lcd(lcd_kontrast2,1); // Kontrast Set C3,C2,C1,C0 zur LCD bertragen
}

// ***********************************************************************
// LCD-Kontrast Einstellungen
// ***********************************************************************
void lcd_kontrast(void) {
byte taste;
  clear_lcd();
  printf_lcd(0,0,"Kontrast:",1);
  printf_lcd(0,1," SET+",1);
  printf_lcd(0,2," SET-",1);
  while(1) {
    // Anzeige der Einstellung
    printf_lcd(7,1,"+",1);
    sprintf_lcd(&str0[0],awd.kontrast,1);
    printf_lcd(8,1,str0,0);
    blink_lcd(8,1,1);    // aktuelle Einstellposition blinken
    taste = get_taste(); // Tastendruck abwarten
    delay(50);
    if (taste == 2) if (awd.kontrast > 1)  awd.kontrast--; // Kontrast-Wert verkleinern
    if (taste == 3) if (awd.kontrast < 30) awd.kontrast++; // Kontrast-Wert vergrern
    if ((taste == 1) || (taste == 4) || (taste == 5)) break; // ESC oder ET oder Timeout
    // LCD Kontrast aktualisieren
    lcd_kontrast_set();
  }
  write_data_flash(); // Werte auch im Daten-FLASH speichern
}

// ***********************************************************************
// IR-SETUP Einstellungen
// ***********************************************************************
void ir_setup(void) {
byte index = 0; // 0...4 Zeiger (0=Adresse 1=ESC 2=SET- 3=SET+ 4=ET)
byte temp[5];   // 0...4 Wert   (0=Adresse 1=ESC 2=SET- 3=SET+ 4=ET)
byte code blink_x[] = {5,5,8,11,14}; // Kursorpositionen
byte blink_y = 1;
byte taste;
  // momentane Werte fr nderungszwecke bernehmen
  temp[0] = awd.rc5addr;
  temp[1] = awd.rc5esc;
  temp[2] = awd.rc5setm;
  temp[3] = awd.rc5setp;
  temp[4] = awd.rc5et;
  clear_lcd();
  printf_lcd(0,0,"IR/RC5 Setup:",1);
  printf_lcd(0,1,"Adr:",1);
  printf_lcd(0,2,"Tas:",1);
  while(1) {
    // Anzeige der aktuellen Einstellungen
    sprintf_lcd(&str0[0],temp[0],1);
    printf_lcd(5,1,str0,0);
    sprintf_lcd(&str0[0],temp[1],1);
    printf_lcd(5,2,str0,0);
    sprintf_lcd(&str0[0],temp[2],1);
    printf_lcd(8,2,str0,0);
    sprintf_lcd(&str0[0],temp[3],1);
    printf_lcd(11,2,str0,0);
    sprintf_lcd(&str0[0],temp[4],1);
    printf_lcd(14,2,str0,0);
    blink_lcd(blink_x[index],blink_y,1); // aktuelle Einstellposition blinken
    taste = get_taste();                 // Tastendruck abwarten
    delay(50);
    if (taste == 2) if (temp[index])      temp[index]--; // Wert verkleinern
    if (taste == 3) if (temp[index] < 64) temp[index]++; // Wert vergrern
    if ((taste == 1) || (taste == 5)) break; // ESC oder Timeout
    if (taste == 4) { // Taste ET (Stellenwechsel)
      if (index < 4) index++;  else index   = 0;
      if (!index) blink_y = 1; else blink_y = 2;
    }
  }
  // genderte Werte bernehmen
  awd.rc5addr  = temp[0];
  awd.rc5esc   = temp[1];
  awd.rc5setm  = temp[2];
  awd.rc5setp  = temp[3];
  awd.rc5et    = temp[4];
  write_data_flash(); // Werte auch im Daten-FLASH speichern
}

// ***********************************************************************
// Information
// ***********************************************************************
void information(void) {
  clear_lcd();
  printf_lcd(0,0,"EQ-KH V1.0 10/08",1);
  printf_lcd(0,1,"89LPC922/TDA7317",1);
  printf_lcd(0,2,"M.Kahnt  (c)2008",1);
  get_taste(); // Tastendruck oder Timeout abwarten
  delay(50);
}

// ***********************************************************************
// Bedienprogramm (Hauptmen)
// ***********************************************************************
void bedienung(void) {
byte menu_POS1 = 0;
byte taste;
char code *code menu_TXT1[] = {
/*  0 */ "\xA5 Equalizer",
/*  1 */ "\xA5 Verst\x84rkung",
/*  2 */ "\xA5 Levelanzeige",
/*  3 */ "\xA5 LCD-Kontrast",
/*  4 */ "\xA5 IR-Setup",
/*  5 */ "\xA5 Information" };
  LED1 = 0;
  while(1) {
    clear_lcd();
    printf_lcd(0,0,"EQ-KH Prozessor",1);
    printf_lcd(0,1,"System-Men\x81:",1);
    while(1) {
      printf_lcd(0,2,menu_TXT1[menu_POS1],1);
      blink_lcd(0,2,1);
      taste = get_taste();
      delay(50);
      if ((taste == 1) || (taste == 4) || (taste == 5)) break; // ESC oder ET oder Timeout
      if (taste == 3) if (menu_POS1 < 5) {menu_POS1++;} else menu_POS1 = 0;
      if (taste == 2) if (menu_POS1 > 0) {menu_POS1--;} else menu_POS1 = 5;
    };
    blink_lcd(0,2,0);
    if ((taste == 1) || (taste == 5)) break; // ESC oder Timeout
    switch (menu_POS1) {
      case 0: // Equalizer
                 equalizer();
                 break;
      case 1: // Verstrkung/Pegel
                 pegel();
                 break;
      case 2: // Levelanzeige
                 level();
	         break;
      case 3: // Kontrast
                 lcd_kontrast();
	         break;
      case 4: // IR-Setup
                 ir_setup();
	         break;
      case 5: // Information
                 information();
    }
  }
  clear_lcd();  // Display lschen
  LED1 = 1;
}

// ***********************************************************************
// Grundinitialisierung des LCD-Moduls in SPI-Mode (seriell)
// ***********************************************************************
void init_lcd(void) {
  // Grundinitialisierung (SPI, wie im 8-Bit parallel-Mode)
  delay(50);         // Pause 50ms nach Power ON
  write_lcd(0x39,1); // Function Set
  write_lcd(0x39,1); // Function Set (gleiches Byte nochmal senden)
  write_lcd(0x15,1); // Bias Set
  write_lcd(0x55,1); // Power Control + Kontrast Set C5,C4
  write_lcd(0x6E,1); // Follower Control
  write_lcd(0x70,1); // Kontrast Set C3,C2,C1,C0
  write_lcd(0x0C,1); // Display Set
  write_lcd(0x06,1); // Entry Mode Set
  clear_lcd();       // Display lschen
  // Power Control und Kontrast Set zwischenspeichen
  // die Angaben werden spter bei der manuellen Kontrasteinstellung wieder gebraucht
  lcd_kontrast1 = 0x55; // Kontrast-Bit C5,C4
  lcd_kontrast2 = 0x70; // Kontrast-Bit C3,C2,C1,C0
}

// ********************************************************************
// Grund-Initialisierung
// ********************************************************************
void init_system (void) {
  delay(50);   // Pause 50ms nach Power ON
  // Grundfunktionen der Portpins setzen
  P0M1 = 0x00; // P0 auf Bidirektional
  P0M2 = 0x00;
  P1M1 = 0x00; // P1 auf Bidirektional
  P1M2 = 0x00;
  P3M1 = 0x00; // P3 auf Bidirektional
  P3M2 = 0x00;
  P0M1 = 0x34; // P0.5 P0.4 P0.2 (0011 0100) auf Input Only (REF / POL / POR)
  P0M2 = 0x80; // P0.7 (1000 0000) als Push-Pull Output (PWM)
  // LED schalten
  LED1    = 1; // LED1 aus
  LED2    = 1; // LED2 aus
  hgl_counter = HGL_ZEIT; // LCD-Hintergrundbeleuchtung Automatik Ein
  LCD_HGL = 1; // LCD-Hintergrundbeleuchtung Ein
  // Timer 0 (Universal-Funktionen) initialisieren
  TMOD = 0x01;      // 0000 0001B    Timer 0: Mode 1 (16-Bit Zhler)
  TH0  = TH0Reload; // Reloadwert Timer 0
  TL0  = TL0Reload;
  TR0  = 1;         // Timer 0 starten
  // Timer 1 (PWM-Funktion) initialisieren
  TMOD  &= 0x0F;    // Timer 1 in Mode 6 (PWM) setzen
  TMOD  |= 0x20;
  TAMOD &= 0xEF;
  TAMOD |= 0x10;
  TH1    = 0x80;    // Anfangs Dutty-Wert 50%
  TL1    = 0x80;
  AUXR1 |= 0x20;    // Enable Toggle Output fr P0.7 (T1)
  TR1    = 1;       // Timer 1 starten
  // Analog-Komparatoren
  PT0AD = 0x34; // P0.5 P0.4 P0.2 (0011 0100) auf Analog Input (REF / POL / POR)
  CMP1  = 0x20; // Komparator funktion fr Kanal 1 freigeben
  CMP2  = 0x20; // Komparator funktion fr Kanal 2 freigeben
  // Interruptsystem
  EX1  = 1;      // Externer Interrupt P1.4
  ET0  = 1;      // Timer 0, Interruptfreigabe
  EA   = 1;      // generelle Interruptfreigabe
  // Werte aus Flash zurcklesen
  awd.kennb = 0;
  read_data_flash(); // Flash-Werte lesen und bernehmen
  if ((awd.kennb != 0x19) || !TASTE_ET){
    // Erstinitialisierung bzw. auf Wunsch falls Taste ET gedrckt war
    awd.pegel    = INIT_PEGEL;
    awd.band1    = INIT_BAND1;
    awd.band2    = INIT_BAND2;
    awd.band3    = INIT_BAND3;
    awd.band4    = INIT_BAND4;
    awd.band5    = INIT_BAND5;
    awd.level    = INIT_LEVEL; 
    awd.kontrast = INIT_KONTRAST;
    awd.rc5addr  = INIT_RC5_ADDR;
    awd.rc5esc   = INIT_RC5_ESC;
    awd.rc5setm  = INIT_RC5_SETM;
    awd.rc5setp  = INIT_RC5_SETP;
    awd.rc5et    = INIT_RC5_ET;
  }
  // TDA 7317 Pegel und Band-Werte setzen
  I2C_start();           // Startsequenz
  I2C_write(ADD_7317);   // Baustein fr schreiben adressieren
  I2C_write(awd.pegel);  // Pegel
  I2C_write(awd.band1);  // Band 1
  I2C_write(awd.band2);  // Band 2
  I2C_write(awd.band3);  // Band 3
  I2C_write(awd.band4);  // Band 4
  I2C_write(awd.band5);  // Band 5
  I2C_stop();            // Stoppsequenz
}

// ***********************************************************************
// Pegelmessung
// ***********************************************************************
void pegelmessung(void) {
  TH1 = 0xFF;         // PWM 0V ausgeben
  delay(10);          // Entladezeit des Kondensators abwarten
  mess_pol = 1;       // Messbedingungen rcksetzen
  mess_por = 1;
  ad_pol   = 0;
  ad_por   = 0;
  mess_zyklus = 1;    // Messzyklus der Komparatoren starten
  while(mess_zyklus); // auf Messergebnis warten (dauert ca. 15ms, max. 50ms bei Vollausstuerung)
}


// ********************************************************************
// Hauptprogramm
// ********************************************************************
void main (void) {
  init_system();      // System initialisierungen
  init_lcd();         // LCD initialisieren
  lcd_kontrast_set(); // LCD Kontrast setzen
  write_lcd_cg_block(sonderzeichen2);   // Sonderzeichen definieren
  printf_lcd(0,0,"EQ-Prozessor RC5",1); // Einschaltmeldung
  printf_lcd(0,1,"V1.0 16.10.2008",1);  
  printf_lcd(0,2,"SDCC-Version",1);  
  delay(2500);  // Pause
  clear_lcd();  // Display lschen
  while(1) {    // Haupschleife
    printf_lcd(0,0,"\xA5\xA5\x45Q-Prozessor\xA5\xA5",1);
    printf_lcd(0,1,"L\x7E",0);
    printf_lcd(0,2,"R\x7E",0);
    write_byte_lcd(14,2,6); // E als Sonderzeichen
    write_byte_lcd(15,2,7); // T als Sonderzeichen
    if (keypressed() > 3) bedienung(); // Bedien-Men, wenn Taste ET oder IR-Taste
    pegelmessung();                 // Level messen
    sprintf_lcd(&str0[0],ad_pol,4); // Level links als Bandanzeige formatieren
    sprintf_lcd(&str0[8],ad_por,4); // Level rechts als Bandanzeige formatieren
    printf_lcd(2,1,&str0[0],0);     // Bandanzeige links anzeigen
    printf_lcd(2,2,&str0[8],0);     // Bandanzeige rechts anzeigen
  }
}
